/*
   This program uses message passing (in this case MPI) to
   implement a synchronization pattern. The message passing
   "primitives" (send & receive) are used as synchronization
   primitives (instead of mutexes, semaphores, or condition variables).

   The pattern is a round-robin ring of working processes.
   Each process "works" for a while and then signals its
   neighbor to begin its turn to work. A process only works
   when it is its turn.

   As an exercise, you should replace the processes with threads
   and then implement this synchronization pattern, first using
   semaphores, and then using condition variables.
*/
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "mpi.h"
#include <windows.h>  // used for Sleep() only

int main(int argc, char** argv)
{
   int my_rank;  /* Rank of process */
   int p;        /* Number of processes */

   int namelen;
   char name[MPI_MAX_PROCESSOR_NAME];
   int tag = 1;       /* Tag for messages */
   MPI_Status status; /* Return status for receive */

   int count = 0;
   int leftSide, rightSide, workTime;

   MPI_Init(&argc, &argv);
   MPI_Comm_rank(MPI_COMM_WORLD, &my_rank);
   MPI_Comm_size(MPI_COMM_WORLD, &p);
   MPI_Get_processor_name(name, &namelen);

   /* compute the rank of my "left-hand" neighbor */
   leftSide = my_rank - 1;
   if (leftSide == -1) leftSide = p-1;
   /* compute the rank of my "right-hand" neighbor */
   rightSide = my_rank + 1;
   if (rightSide == p) rightSide = 0;

   srand(my_rank + time(NULL));  /* seed the random number generator */

   if (my_rank == 0)
   {
      while(count < 20)
      {
         /* Do some "work" */
         count++;
         workTime = 1+ rand()%5;
         printf("%d: Doing %d seconds of work by rank %d of %d on %s...", count, workTime, my_rank, p, name);
         Sleep(workTime*1000);  // Win32
         //sleep(workTime);     // Posix
         printf("...Done\n");

         /* Send a message to my right-hand neighbor */
         MPI_Send(&count, 1, MPI_INT, rightSide, tag, MPI_COMM_WORLD);

         /* Wait for a message from my left-hand neighbor */
         MPI_Recv(&count, 1, MPI_INT, leftSide, tag, MPI_COMM_WORLD, &status);
      }
   }
   else /* my_rank != 0 */
   {
      while(count < 20)
      {
         /* Wait for a message from my left-hand neighbor */
         MPI_Recv(&count, 1, MPI_INT, leftSide, tag, MPI_COMM_WORLD, &status);

         /* Do some "work" */
         count++;
         workTime = 1+ rand()%5;
         printf("%d: Doing %d seconds of work by rank %d of %d on %s...", count, workTime, my_rank, p, name);
         Sleep(workTime*1000);  // Win32
         //sleep(workTime);     // Posix
         printf("...Done\n");

         /* Send a message to my right-hand neighbor */
         MPI_Send(&count, 1, MPI_INT, rightSide, tag, MPI_COMM_WORLD);
      }
   }

   MPI_Finalize();
   return 0;
}
